{$m+} //generate of runtime type information (RTTI) for user defined types
unit JsonUserDefinedTypes;

interface

uses
  JsonStdTypes, SynCommons, mORMot, Classes;

type

  TUserRecord= packed record //packed only!!!
    rint: integer;
    ri64: int64;
    rdub: double;
    rdt : TDateTime;
    rstr: string;
    end;
  TDynRecordArray= array of TUserRecord;

  TUserClass = class(TObject)
  public
    fint: integer;
    fi64: int64;
    fdub: double;
    fdt: TDateTime;
    fstr: string;
    fIntArr: TDynIntegerArray;
    fI64Arr: TDynInt64Array;
    fDubArr: TDynDoubleArray;
    fDTArr : TDynDateTimeArray;
    fStrArr: TDynStringArray;
    fRecArr: TDynRecordArray;
  published
    property int: integer read fint write fint;
    property i64: int64 read fi64 write fi64;
    property dub: double read fdub write fdub;
    property dt : TDateTime read fdt write fdt;
    property str: string read fstr write fstr;
    property IntArr: TDynIntegerArray read fIntArr write fIntArr;
    property I64Arr: TDynInt64Array read fI64Arr write fI64Arr;
    property DubArr: TDynDoubleArray read fDubArr write fDubArr;
    property DTArr : TDynDateTimeArray read fDTArr write fDTArr;
    property StrArr: TDynStringArray read fStrArr write fStrArr;
    property RecArr: TDynRecordArray read fRecArr write fRecArr;
  end;

procedure SetJsonRecordArrayFormat(aFormat: integer; aUserInfo: pointer= nil);

implementation

const
  NULCHARS: array[0..1] of AnsiChar= #0#0; //two nulls to allow increment of pointer to the first of them

//----------------------------------------------------------------------------------------------------------------------

procedure RecordArrayWriterNoNames(aUserInfo: pointer; const aWriter: TTextWriter; const aValue);
var
  v: TUserRecord absolute aValue;
begin;
  aWriter.Add('[%,%,%,"%","%"]',
    [v.rint, v.ri64, v.rdub, DateTimeToIso8601Text(v.rdt), StringToUTF8(v.rstr)],
    twJSONEscape);
  end;

function RecordArrayReaderNoNames(aUserInfo: pointer; p: PUTF8Char; var aValue; out aValid: Boolean): PUTF8Char;
var
  v: TUserRecord absolute aValue;
  wasString: boolean;
  err: integer;
  delim: AnsiChar;
begin;
  aValid:=false;
  Result:=nil;
  if (p=nil) or (p^<>'[') then exit;
  inc(p);
  v.rint:=GetInteger(GetJSONField(p,p,@wasString,@delim));                   if (delim<>',') or wasString then exit;
  v.ri64:=GetInt64(GetJSONField(p,p,@wasString,@delim));                     if (delim<>',') or wasString then exit;
  v.rdub:=GetExtended(GetJSONField(p,p,@wasString,@delim),err);              if (delim<>',') or wasString or (err<>0) then exit;
  v.rdt :=Iso8601ToDateTimePUTF8Char(GetJSONField(p,p,@wasString,@delim));   if (delim<>',') or not wasString then exit;
  v.rstr:=UTF8ToString(GetJSONField(p,p,@wasString,@delim));                 if (delim<>']') or not wasString then exit;
  if p=nil then p:=@NULCHARS; //end of text is not error
  aValid:=true;
  Result:=p; // ',' or ']' for last item of array
  end;

//----------------------------------------------------------------------------------------------------------------------

procedure RecordArrayWriterWithNames(aUserInfo: pointer; const aWriter: TTextWriter; const aValue);
var
  v: TUserRecord absolute aValue;
begin;
  aWriter.AddJSONEscape(['rint', v.rint,
                         'ri64', v.ri64,
                         'rdub', v.rdub,
                         'rdt' , DateTimeToIso8601Text(v.rdt),
                         'rstr', StringToUTF8(v.rstr)]);
  end;

function RecordArrayReaderWithNames(aUserInfo: pointer; p: PUTF8Char; var aValue; out aValid: Boolean): PUTF8Char;
var
  v: TUserRecord absolute aValue;
  Values: TPUtf8CharDynArray;
  err: integer;
begin;
  aValid:=false;
  if (p=nil) or (p^<>'{') then exit;
  Result:=JSONDecode(p,['rint','ri64','rdub', 'rdt', 'rstr'],Values);
  if Result=nil then exit; // Result^ = ',' or ']' for last item of array
  v.rint:=GetInteger(Values[0]);
  v.ri64:=GetInt64(Values[1]);
  v.rdub:=GetExtended(Values[2],err); if err<>0 then exit;
  v.rdt :=Iso8601ToDateTimePUTF8Char(Values[3]);
  v.rstr:=UTF8DecodeToString(Values[4],StrLen(Values[4]));
  aValid:=(Values[0]<>nil) and (Values[1]<>nil) and (Values[2]<>nil) and (Values[3]<>nil) and (Values[4]<>nil);
  end;

//----------------------------------------------------------------------------------------------------------------------

procedure SetJsonRecordArrayFormat(aFormat: integer; aUserInfo: pointer= nil);
begin;
  case aFormat of
    1: begin; //array of values = "name":[["value1","value2"],...]
       RegisterDynArrayRW(TypeInfo(TDynRecordArray), RecordArrayReaderNoNames, RecordArrayWriterNoNames, aUserInfo);
       end;
    2: begin; //set of AV-pairs = "name":[{"name1":"value1","name2":"value2"},...]
       RegisterDynArrayRW(TypeInfo(TDynRecordArray), RecordArrayReaderWithNames, RecordArrayWriterWithNames, aUserInfo);
       end;
    else begin; //0: Base64 (default) = "name":["base64"]
       RegisterDynArrayRW(TypeInfo(TDynRecordArray), nil, nil);
       end;
    end;
  end;

//----------------------------------------------------------------------------------------------------------------------

initialization

  TJSONSerializer.RegisterClassForJSON(TUserClass); //or [class1,class2,...] need for JsonToObject(ObjectList,...)

end.


